"""Gen triple-set from a  matrix."""
# pylint: disable=unused-import

from typing import List, Tuple, Union  # noqa

import numpy as np
import pandas as pd


# fmt: off
def cmat2tset(
    cmat1: Union[List[List[float]], np.ndarray, pd.DataFrame],
    # thirdcol: bool = True
) -> np.ndarray:
    # ) -> List[Union[Tuple[int, int], Tuple[int, int, float]]]:
    # fmt: on
    """Gen triple-set from a matrix.

    Args
        cmat: 2d-array or list, correlation or other metric matrix
        # thirdcol: bool, whether to output a third column (max value)

    Returns
        Obtain the max and argmax for each column, erase the row afterwards to eliminate one single row  that would dominate
        every column.
    """
    # if isinstance(cmat, list):
    cmat = np.array(cmat1)

    if not np.prod(cmat.shape):
        raise SystemError("data not 2d...")

    _ = """
    # y00 = range(cmat.shape[1])  # cmat.shape[0] long time wasting bug

    yargmax = cmat.argmax(axis=0)
    if thirdcol:
        ymax = cmat.max(axis=0)

        res = [*zip(y00, yargmax, ymax)]  # type: ignore
        # to unzip
        # a, b, c = zip(*res)

        return res

    _ = [*zip(y00, yargmax)]  # type: ignore
    return _
    """
    low_ = cmat.min() - 1
    argmax_max = []
    src_len, tgt_len = cmat.shape  # ylim, xlim
    for _ in range(min(src_len, tgt_len)):
        argmax = int(cmat.argmax())
        row, col = divmod(argmax, tgt_len)
        argmax_max.append([col, row, cmat.max()])  # x-axis, y-axis

        # erase row-th row and col-th col of cmat
        cmat[row, :] = low_
        cmat[:, col] = low_

    return np.array(argmax_max)



"""Genereat align set (aset) based on pset (pair set), src_lang and tgt_len."""
# pylint: disable=unused-variable

from typing import List, Tuple, Union
from itertools import zip_longest

# from logzero import logger


# fmt: off
def gen_aset(
        pset: List[Tuple[Union[str, float], Union[str, float], Union[str, float]]],
        src_len: int,  # n_rows
        tgt_len: int,  # n_cols
) -> List[Tuple[Union[str, float], Union[str, float], Union[str, float]]]:
    # fmt: on
    """Genereat align set (aset) based on pset, src_lang and tgt_len.

    src_len, tgt_len = cmat.shape
    zip_longest(..., fillvalue="")

    Args:
        pset: [x(lang2 zh), y(lang1 en), cos]
        src_len: lang1 (en)
        tgt_len: lang2 (zh)

    Returns:
        aset:
        [0...tgt_len, 0...src_len]
        [0, 0, .]
        ...
        [tgt_len-1, src_len-1, .]
    """
    # empty pset []
    if not pset:
        return [*zip_longest(range(tgt_len), range(src_len), fillvalue="")]
    # empty [[]]
    if len(pset) == 1:
        if not pset[0]:
            return [*zip_longest(range(tgt_len), range(src_len), fillvalue="")]

    buff = []
    pos0, pos1 = -1, -1
    for elm in pset:
        # elm0, elm1, elm2 = elm
        elm0, elm1, *elm2 = elm
        elm0 = int(elm0)
        elm1 = int(elm1)
        interval = max(elm0 - pos0 - 1, elm1 - pos1 - 1)
        _ = zip_longest(range(pos0 + 1, elm0), range(pos1 + 1, elm1), [""] * interval, fillvalue="")
        buff.extend(_)
        buff.append(elm)
        pos0, pos1 = elm0, elm1

    # last batch if any
    elm0, elm1 = tgt_len, src_len
    interval = max(elm0 - pos0 - 1, elm1 - pos1 - 1)
    _ = zip_longest(range(pos0 + 1, elm0), range(pos1 + 1, elm1), [""] * interval, fillvalue="")
    buff.extend(_)

    return buff



"""Gne pset from cmat. Find pairs for a given cmat.

tinybee.find_pairs.py with fixed estimator='dbscan' eps=eps, min_samples=min_samples
"""
# pylint: disable=too-many-locals, unused-import, invalid-name

from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
import logzero
from logzero import logger
from cmat2aset.cmat2tset import cmat2tset
from cmat2aset.interpolate_pset import interpolate_pset


def _gen_pset(
    cmat1: Union[List[List[float]], np.ndarray, pd.DataFrame],
    eps: float = 10,
    min_samples: int = 6,
    delta: float = 7,
    verbose: Union[bool, int] = False,
    # ) -> List[Tuple[int, int, Union[float, str]]]:
) -> List[Tuple[Union[float, str], Union[float, str], Union[float, str]]]:
    """Gen pset from cmat.

    Find pairs for a given cmat.

    Args:
        cmat: correlation/similarity matrix
        eps: min epsilon for DBSCAN (10)
        min_samples: minimum # of samples for DBSCAN (6)
        delta: tolerance (7)

    Returns:
        pairs + "" or metric (float)

    dbscan_pairs' setup
        if eps is None:
            eps = src_len * .01
            if eps < 3:
                eps = 3
        if min_samples is None:
            min_samples = tgt_len / 100 * 0.5
            if min_samples < 3:
                min_samples = 3

    def gen_eps_minsamples(src_len, tgt_len):
        eps = src_len * .01
        if eps < 3:
            eps = 3

        min_samples = tgt_len / 100 * 0.5
        if min_samples < 3:
            min_samples = 3
        return {"eps": eps, "min_samples": min_samples}

    """
    if isinstance(verbose, bool):
        if verbose:
            verbose = 10
        else:
            verbose = 20
    logzero.loglevel(verbose)

    # if isinstance(cmat, list):
    cmat = np.array(cmat1)

    src_len, tgt_len = cmat.shape

    # tset = cmat2tset(cmat)
    tset = cmat2tset(cmat).tolist()

    logger.debug("tset: %s", tset)

    # iset = gen_iset(cmat, verbose=verbose, estimator=estimator)
    labels = DBSCAN(eps=eps, min_samples=min_samples).fit(tset).labels_

    df_tset = pd.DataFrame(tset, columns=["x", "y", "cos"])
    cset = df_tset[labels > -1].to_numpy()

    # sort cset
    _ = sorted(cset.tolist(), key=lambda x: x[0])
    iset = interpolate_pset(_, tgt_len)

    # *_, ymax = zip(*tset)
    # ymax = list(ymax)
    # low_ = np.min(ymax) - 1  # reset to minimum_value - 1

    buff = [(-1, -1, ""), (tgt_len, src_len, "")]

    # for idx, tset_elm in enumerate(tset):
    for tset_elm in tset:
        logger.debug("buff: %s", buff)
        # postion max in ymax and insert in buff
        # if with range given by iset+-delta and
        # it's valid (do not exceed constraint
        # by neighboring points

        # argmax = int(np.argmax(ymax))

        # logger.debug("=== %s,%s === %s", _, argmax, tset[_])
        logger.debug("=== %s === %s", _, tset_elm)

        # ymax[_] = low_
        # elm = tset[argmax]
        # elm0, *_ = elm

        elm0, *_ = tset_elm

        # position elm in buff
        idx = -1  # for making pyright happy
        for idx, loc in enumerate(buff):
            if loc[0] > elm0:
                break
        else:
            idx += 1  # last

        # insert elm in for valid elm
        # (within range inside two neighboring points)

        # pos = int(tset[argmax][0])
        pos = int(tset_elm[0])
        logger.debug(" %s <=> %s ", tset_elm, iset[pos])

        # if abs(tset[argmax][1] - iset[pos][1]) <= delta:
        if abs(tset_elm[1] - iset[pos][1]) <= delta:
            if tset_elm[1] > buff[idx - 1][1] and tset_elm[1] < buff[idx][1]:
                buff.insert(idx, tset_elm)
                logger.debug("idx: %s, tset_elm: %s", idx, tset_elm)
            else:
                logger.debug("\t***\t idx: %s, tset_elm: %s", idx, tset_elm)
        _ = """
        if abs(tset[loc][1] - iset[loc][1]) <= delta:
            if tset[loc][1] > buff[idx][1] and tset[loc][1] < buff[idx + 1][1]:
                buff.insert(idx + 1, tset[loc])
        # """

    # remove first and last entry in buff
    buff.pop(0)
    buff.pop()

    # return [(1, 1, "")]
    return [(int(elm0), int(elm1), elm2) for elm0, elm1, elm2 in buff]


def gen_pset(
    cmat1: Union[List[List[float]], np.ndarray, pd.DataFrame],
    eps: float = 10,
    min_samples: int = 6,
    delta: float = 7,
    verbose: Union[bool, int] = False,
) -> List[Tuple[Union[float, str], Union[float, str], Union[float, str]]]:
    """Gen pset.

    Refer to _gen_pset.
    """
    del verbose
    gen_pset.min_samples = min_samples
    for min_s in range(min_samples):
        logger.debug(" min_samples, try %s", min_samples - min_s)
        try:
            pset = _gen_pset(
                cmat1,
                eps=eps,
                min_samples=min_samples - min_s,
                delta=delta,
            )
            break
        except ValueError:
            logger.debug(" decrease min_samples by %s", min_s + 1)
            continue
        except Exception as e:
            logger.error(e)
            continue
    else:
        # break should happen above when min_samples = 2
        raise Exception("bummer, this shouldn't happen, probably another bug")

    # store new min_samples
    gen_pset.min_samples = min_samples - min_s

    return pset



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cmat2aset']

package_data = \
{'': ['*']}

install_requires = \
['icecream>=2.1.1,<3.0.0',
 'install>=1.3.5,<2.0.0',
 'logzero>=1.7.0,<2.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'sklearn>=0.0,<0.1']

setup_kwargs = {
    'name': 'cmat2aset',
    'version': '0.1.0a6',
    'description': 'correlation matrix to align set ',
    'long_description': '# cmat2aset\n[![pytest](https://github.com/ffreemt/cmat2aset/actions/workflows/routine-tests.yml/badge.svg)](https://github.com/ffreemt/cmat2aset/actions)[![python](https://img.shields.io/static/v1?label=python+&message=3.8%2B&color=blue)](https://www.python.org/downloads/)[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)[![PyPI version](https://badge.fury.io/py/cmat2aset.svg)](https://badge.fury.io/py/cmat2aset)\n\ncmat2aset descr\n\n## Install it\n\n```shell\npip install git+https://github.com/ffreemt/cmat2aset\n# poetry add git+https://github.com/ffreemt/cmat2aset\n# git clone https://github.com/ffreemt/cmat2aset && cd cmat2aset\n```\n\n## Use it\n```python\nfrom cmat2aset import cmat2aset\n\n```\n',
    'author': 'ffreemt',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ffreemt/cmat2aset',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.3,<4.0',
}


setup(**setup_kwargs)



"""Interpolate np.nan."""
# pylint: disable=invalid-name
from typing import List, Tuple
import numpy as np
import pandas as pd


# fmt: off
def interpolate_pset(
        pairs: List[Tuple[int, int, float]],
        tgt_len: int,
        method: str = 'linear',
        limit_direction: str = 'both',
) -> List[Tuple[int, int]]:
    # fmt: on
    """Interpolate.

    Args:
        pairs: integer pairs, some np.nan
        tgt_len: over 0...tgt_len-1 (x-axis, cmat.shape[1])
        method: for use in pd.DataFrame.interpolate
        limit_direction:  for use in pd.DataFrame.interpolate
    Returns:
        np.nan converted
    """
    y00, *_ = zip(*pairs)

    res = []
    for idx in range(tgt_len):
        if idx in y00:
            loc = y00.index(idx)
            res.append(tuple(pairs[loc][:2]))
        else:
            res.append((idx, np.nan))

    df = pd.DataFrame(res, columns=["y00", "yargmax"])
    _ = df.interpolate(method=method, limit_direction=limit_direction, axis=0)

    _ = _.to_numpy(dtype=int)
    _ = [(int(elm0), int(elm1)) for elm0, elm1 in _]

    return _
